import { Body as Content } from 'aws-sdk/clients/s3';
import { DownloadCenterConfig } from './download-center-config';
export declare type S3BucketConfig = {
    /**
     * The bucket name.
     */
    bucket: string;
    /**
     * The AWS access key id.
     */
    accessKeyId: string;
    /**
     * The AWS secret access key.
     */
    secretAccessKey: string;
    /**
     * S3 service endpoint. Set this to connect to a local test server.
     */
    endpoint?: string;
    /**
     * Whether to force path style URLs for S3 objects..
     *
     * The default is false. Set this to `true`
     * to connect to a local test server using an arbitrary endpoint.
     */
    s3ForcePathStyle?: boolean;
    /**
     * Whether or not TLS should be enabled.
     *
     * The default is `true`. Set this to `false`
     * to connect to a local test server.
     */
    sslEnabled?: boolean;
};
export declare type UploadAssetOptions = {
    contentType?: string;
};
declare type DownloadCenterConfigPlatform = {
    [key: string]: any;
    download_link: string;
};
declare type ProbeResponse = {
    ok: boolean;
    status: number;
};
/**
 * Vaidates a download center configuration object against a json schema.
 * Throws an error if the configuration is invalid.
 *
 * @static
 * @param {DownloadCenterConfig} config -
 *  the download center product configuration document.
 * @memberof DownloadCenter
 */
export declare function validateConfigSchema(config: DownloadCenterConfig): void;
/**
 * Validates all the asset links referenced in a configuration object.
 * Makes an HEAD http call for each asset link and throws an error in case
 * an asset is not reacheable.
 *
 * @static
 * @param {DownloadCenterConfig} config -
 *  the download center product configuration document.
 * @return {Promise<void>}
 * @memberof DownloadCenter
 */
export declare function validateDownloadLinks(config: DownloadCenterConfig): Promise<void>;
/**
 * Probes the download_link for a configuration platform.
 * Returns the response of the probe.
 *
 * @static
 * @param {DownloadCenterConfigPlatform} platform
 * @return {Promise<ProbeResponse>}
 * @memberof DownloadCenter
 */
export declare function probePlatformDownloadLink(platform: DownloadCenterConfigPlatform): Promise<ProbeResponse>;
/**
 * Validates a download center config object.
 * Throws an error if the object has an incorrect format or if any of the
 * assets link is not reachable.
 *
 * @static
 * @param {DownloadCenterConfig} config -
 *  the download center product configuration document.
 * @return {Promise<void>}
 * @memberof DownloadCenter
 */
export declare function validateConfig(config: DownloadCenterConfig): Promise<void>;
export declare class DownloadCenter {
    private s3Upload;
    private s3GetObject;
    private s3BucketName;
    constructor(bucketConfig: S3BucketConfig);
    /**
     * Downloads an asset from the download center. This is equivalent to a
     * get object from the download center s3 bucket.
     *
     * @param {string} s3ObjectKey -
     *  the s3 object key of the asset that has to be downloaded,
     *  ie. `my-project/asset.zip`.
     * @return {(Promise<Content | undefined>)}
     * @memberof DownloadCenter
     */
    downloadAsset(s3ObjectKey: string): Promise<Content | undefined>;
    /**
     * Uploads an asset to the download center. This is equivalent to an
     * upload to the download center s3 bucket.
     *
     * @param {string} s3ObjectKey -
     *  the s3 object key of the asset that has to be uploaded,
     *  ie. `my-project/asset.zip`.
     * @param {Content} content -
     *  a string, Buffer, Uint8Array, Blob or Readable containing the data of
     *  the asset to be uploaded.
     * @param {UploadAssetOptions} [options={}] -
     *  metadata for the upload.
     * @param {string} [options.contentType=undefined] -
     *  an optional content type of the asset. If not specified
     *  will be detected by s3.
     * @return {Promise<void>}
     * @memberof DownloadCenter
     */
    uploadAsset(s3ObjectKey: string, content: Content, options?: UploadAssetOptions): Promise<void>;
    /**
     * Downloads a product configuration from the download center bucket.
     *
     * @param {string} s3ObjectKey  -
     *  the s3 object key of the configuration that has to be downloaded,
     *  ie. `products/compass.json`.
     * @return {(Promise<DownloadCenterConfig | undefined>)} -
     *  the download center product configuration document.
     * @memberof DownloadCenter
     */
    downloadConfig(s3ObjectKey: string): Promise<DownloadCenterConfig | undefined>;
    /**
     * Validates and uploads a product configuration document
     * to the download center s3 bucket.
     *
     * @param {string} s3ObjectKey  -
     *  the s3 object key of the configuration that has to be uploaded,
     *  ie. `products/compass.json`.
     * @param {DownloadCenterConfig} config -
     *  the download center product configuration document.
     *
     * @return {Promise<void>}
     * @memberof DownloadCenter
     */
    uploadConfig(s3ObjectKey: string, config: DownloadCenterConfig): Promise<void>;
}
export {};
