"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const mongodb_redact_1 = __importDefault(require("mongodb-redact"));
const history_1 = require("@mongosh/history");
const util_1 = require("util");
const build_info_1 = require("./build-info");
const log_writer_1 = require("./log-writer");
class NoopAnalytics {
    identify(_info) { }
    track(_info) { }
}
function setupLoggerAndTelemetry(logId, bus, makeLogger, makeAnalytics) {
    const log = makeLogger();
    const mongosh_version = require('../package.json').version;
    let userId;
    let telemetry;
    const userTraits = {
        platform: process.platform,
        arch: process.arch
    };
    log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000000), 'log', 'Starting log', {
        execPath: process.execPath,
        ...(0, build_info_1.buildInfo)()
    });
    let analytics = new NoopAnalytics();
    try {
        analytics = makeAnalytics();
    }
    catch (e) {
        log.error('MONGOSH', (0, log_writer_1.mongoLogId)(1000000001), 'analytics', 'Failed to instantiate analytics provider', e);
    }
    let hasStartedMongoshRepl = false;
    bus.on('mongosh:start-mongosh-repl', (ev) => {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000002), 'repl', 'Started REPL', ev);
        hasStartedMongoshRepl = true;
    });
    let usesShellOption = false;
    bus.on('mongosh:start-loading-cli-scripts', (event) => {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000003), 'repl', 'Start loading CLI scripts');
        usesShellOption = event.usesShellOption;
    });
    bus.on('mongosh:connect', function (args) {
        const connectionUri = (0, history_1.redactURICredentials)(args.uri);
        const { uri: _uri, ...argsWithoutUri } = args;
        const params = { session_id: logId, userId, connectionUri, ...argsWithoutUri };
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000004), 'connect', 'Connecting to server', params);
        if (telemetry) {
            analytics.track({
                userId,
                event: 'New Connection',
                properties: {
                    mongosh_version,
                    session_id: logId,
                    ...argsWithoutUri
                }
            });
        }
    });
    bus.on('mongosh:driver-initialized', function (driverMetadata) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000004), 'connect', 'Driver initialized', driverMetadata);
    });
    bus.on('mongosh:new-user', function (id, enableTelemetry) {
        userId = id;
        telemetry = enableTelemetry;
        if (telemetry)
            analytics.identify({ userId, traits: userTraits });
    });
    bus.on('mongosh:update-user', function (id, enableTelemetry) {
        userId = id;
        telemetry = enableTelemetry;
        if (telemetry)
            analytics.identify({ userId, traits: userTraits });
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000005), 'config', 'User updated', { enableTelemetry });
    });
    bus.on('mongosh:error', function (error, context) {
        if (context === 'fatal') {
            log.fatal('MONGOSH', (0, log_writer_1.mongoLogId)(1000000006), context, `${error.name}: ${error.message}`, error);
        }
        else {
            log.error('MONGOSH', (0, log_writer_1.mongoLogId)(1000000006), context, `${error.name}: ${error.message}`, error);
        }
        if (telemetry && error.name.includes('Mongosh')) {
            analytics.track({
                userId,
                event: 'Error',
                properties: {
                    mongosh_version,
                    name: error.name,
                    code: error.code,
                    scope: error.scope,
                    metadata: error.metadata
                }
            });
        }
    });
    bus.on('mongosh:evaluate-input', function (args) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000007), 'repl', 'Evaluating input', args);
    });
    bus.on('mongosh:use', function (args) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000008), 'shell-api', 'Used "use" command', args);
        if (telemetry) {
            analytics.track({
                userId,
                event: 'Use',
                properties: {
                    mongosh_version
                }
            });
        }
    });
    bus.on('mongosh:show', function (args) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000009), 'shell-api', 'Used "show" command', args);
        if (telemetry) {
            analytics.track({
                userId,
                event: 'Show',
                properties: {
                    mongosh_version,
                    method: args.method
                }
            });
        }
    });
    bus.on('mongosh:setCtx', function (args) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000010), 'shell-api', 'Initialized context', args);
    });
    bus.on('mongosh:api-call', function (args) {
        let arg;
        try {
            arg = JSON.parse(JSON.stringify(args));
        }
        catch (_a) {
            arg = { _inspected: (0, util_1.inspect)(args) };
        }
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000011), 'shell-api', 'Performed API call', (0, mongodb_redact_1.default)(arg));
    });
    bus.on('mongosh:api-load-file', function (args) {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000012), 'shell-api', 'Loading file via load()', args);
        if (telemetry) {
            analytics.track({
                userId,
                event: hasStartedMongoshRepl ? 'Script Loaded' : 'Script Loaded CLI',
                properties: {
                    mongosh_version,
                    nested: args.nested,
                    ...(hasStartedMongoshRepl ? {} : { shell: usesShellOption })
                }
            });
        }
    });
    bus.on('mongosh:eval-cli-script', function () {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000013), 'repl', 'Evaluating script passed on the command line');
        if (telemetry) {
            analytics.track({
                userId,
                event: 'Script Evaluated',
                properties: {
                    mongosh_version,
                    shell: usesShellOption
                }
            });
        }
    });
    bus.on('mongosh:mongoshrc-load', function () {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000014), 'repl', 'Loading .mongoshrc.js');
        if (telemetry) {
            analytics.track({
                userId,
                event: 'Mongoshrc Loaded',
                properties: {
                    mongosh_version
                }
            });
        }
    });
    bus.on('mongosh:mongoshrc-mongorc-warn', function () {
        log.info('MONGOSH', (0, log_writer_1.mongoLogId)(1000000015), 'repl', 'Warning about .mongorc.js/.mongoshrc.js mismatch');
        if (telemetry) {
            analytics.track({
                userId,
                event: 'Mongorc Warning',
                properties: {
                    mongosh_version
                }
            });
        }
    });
    bus.on('mongosh:mongocryptd-tryspawn', function (ev) {
        log.info('MONGOCRYPTD', (0, log_writer_1.mongoLogId)(1000000016), 'mongocryptd', 'Trying to spawn mongocryptd', ev);
    });
    bus.on('mongosh:mongocryptd-error', function (ev) {
        log.warn('MONGOCRYPTD', (0, log_writer_1.mongoLogId)(1000000017), 'mongocryptd', 'Error running mongocryptd', ev);
    });
    bus.on('mongosh:mongocryptd-log', function (ev) {
        log.info('MONGOCRYPTD', (0, log_writer_1.mongoLogId)(1000000018), 'mongocryptd', 'mongocryptd log message', ev);
    });
    bus.on('mongosh-snippets:loaded', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000019), 'snippets', 'Loaded snippets', ev);
    });
    bus.on('mongosh-snippets:npm-lookup', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000020), 'snippets', 'Performing npm lookup', ev);
    });
    bus.on('mongosh-snippets:npm-lookup-stopped', function () {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000021), 'snippets', 'npm lookup stopped');
    });
    bus.on('mongosh-snippets:npm-download-failed', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000022), 'snippets', 'npm download failed', ev);
    });
    bus.on('mongosh-snippets:npm-download-active', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000023), 'snippets', 'npm download active', ev);
    });
    bus.on('mongosh-snippets:fetch-index', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000024), 'snippets', 'Fetching snippet index', ev);
    });
    bus.on('mongosh-snippets:fetch-cache-invalid', function () {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000025), 'snippets', 'Snippet cache invalid');
    });
    bus.on('mongosh-snippets:fetch-index-error', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000026), 'snippets', 'Fetching snippet index failed', ev);
    });
    bus.on('mongosh-snippets:fetch-index-done', function () {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000027), 'snippets', 'Fetching snippet index done');
    });
    bus.on('mongosh-snippets:package-json-edit-error', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000028), 'snippets', 'Modifying snippets package.json failed', ev);
    });
    bus.on('mongosh-snippets:spawn-child', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000029), 'snippets', 'Spawning helper', ev);
    });
    bus.on('mongosh-snippets:load-snippet', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000030), 'snippets', 'Loading snippet', ev);
    });
    bus.on('mongosh-snippets:snippet-command', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000031), 'snippets', 'Running snippet command', ev);
        if (telemetry && ev.args[0] === 'install') {
            analytics.track({
                userId,
                event: 'Snippet Install',
                properties: {
                    mongosh_version
                }
            });
        }
    });
    bus.on('mongosh-snippets:transform-error', function (ev) {
        log.info('MONGOSH-SNIPPETS', (0, log_writer_1.mongoLogId)(1000000032), 'snippets', 'Rewrote error message', ev);
    });
    const deprecatedApiCalls = new Set();
    bus.on('mongosh:deprecated-api-call', function (ev) {
        deprecatedApiCalls.add(`${ev.class}#${ev.method}`);
    });
    bus.on('mongosh:evaluate-finished', function () {
        deprecatedApiCalls.forEach(e => {
            const [clazz, method] = e.split('#');
            log.warn('MONGOSH', (0, log_writer_1.mongoLogId)(1000000033), 'shell-api', 'Deprecated API call', { class: clazz, method });
            if (telemetry) {
                analytics.track({
                    userId,
                    event: 'Deprecated Method',
                    properties: {
                        mongosh_version,
                        class: clazz,
                        method
                    }
                });
            }
        });
        deprecatedApiCalls.clear();
    });
    bus.on('mongosh-sp:connect-heartbeat-failure', function (ev) {
        var _a;
        log.warn('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000034), 'connect', 'Server heartbeat failure', {
            ...ev,
            failure: (_a = ev.failure) === null || _a === void 0 ? void 0 : _a.message
        });
    });
    bus.on('mongosh-sp:connect-heartbeat-succeeded', function (ev) {
        log.info('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000035), 'connect', 'Server heartbeat succeeded', ev);
    });
    bus.on('mongosh-sp:connect-fail-early', function () {
        log.warn('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000036), 'connect', 'Aborting connection attempt as irrecoverable');
    });
    bus.on('mongosh-sp:connect-attempt-finished', function () {
        log.info('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000037), 'connect', 'Connection attempt finished');
    });
    bus.on('mongosh-sp:resolve-srv-error', function (ev) {
        var _a;
        log.error('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000038), 'connect', 'Resolving SRV record failed', {
            from: (0, history_1.redactURICredentials)(ev.from),
            error: (_a = ev.error) === null || _a === void 0 ? void 0 : _a.message,
            duringLoad: ev.duringLoad
        });
    });
    bus.on('mongosh-sp:resolve-srv-succeeded', function (ev) {
        log.info('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000039), 'connect', 'Resolving SRV record succeeded', {
            from: (0, history_1.redactURICredentials)(ev.from),
            to: (0, history_1.redactURICredentials)(ev.to)
        });
    });
    bus.on('mongosh-sp:reset-connection-options', function () {
        log.info('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000040), 'connect', 'Reconnect because of changed connection options');
    });
    bus.on('mongosh-sp:missing-optional-dependency', function (ev) {
        log.error('MONGOSH-SP', (0, log_writer_1.mongoLogId)(1000000041), 'deps', 'Missing optional dependency', {
            name: ev.name,
            error: ev === null || ev === void 0 ? void 0 : ev.error.message
        });
    });
}
exports.default = setupLoggerAndTelemetry;
//# sourceMappingURL=setup-logger-and-telemetry.js.map